#!/bin/bash -e
#
# prepare-raspbian-sd:bin/api/prepare-raspbian-config-json-api.sh
#
declare -gr __PREPARE_RASPBIAN_CONFIG_DIR__="${HOME}/.config/prepare-raspbian-sd"
declare -gr _PREPARE_RASPBIAN_CONFIG_FILE_="${__PREPARE_RASPBIAN_CONFIG_DIR__}/config.json"
declare -gr _PREPARE_RASPBIAN_CONFIG_VARS_JSON_='{
  "BCM_WORKSPACE":"boolean",
  "DEFAULT_SD_DEV_NAME":"string",
  "GPU_MEM":"number",
  "RASPBIAN_DEFAULT_BACKUP_DIRECTORY":"string",
  "RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY":"string",
  "RASPBIAN_DEFAULT_IMG":"string",
  "RASPBIAN_PI_GROUP_ID":"number",
  "RASPBIAN_PI_USER_ID":"number",
  "SSH_AUTHORIZED_KEYS":"string",
  "SSH_CONFIGURE":"boolean",
  "SSH_ENABLE":"boolean"
}'

function prepare_raspbian_get_config_filename {
  echo "${_PREPARE_RASPBIAN_CONFIG_FILE_}"
}

function prepare_raspbian_get_config_filename_if_exist {
  local filename=
  filename="$( prepare_raspbian_get_config_filename )" || return $?

  if [ -f "${filename}" ] ; then
    echo "${filename}"
  else
    echofatal "Can't find configuration file: '${filename}'"
    return 1
  fi
}

function prepare_raspbian_get_type {
  local -r var_name="$1"

  jq -r ".[\"${var_name}\"]" <<<"${_PREPARE_RASPBIAN_CONFIG_VARS_JSON_}" || return $?
}

function prepare_raspbian_show_filename {
  prepare_raspbian_get_config_filename || return $?
}

function prepare_raspbian_show_list {
  jq -r 'keys[]' <<<"${_PREPARE_RASPBIAN_CONFIG_VARS_JSON_}" || return $?
}

function prepare_raspbian_show_list_json {
  jq -S 'keys' <<<"${_PREPARE_RASPBIAN_CONFIG_VARS_JSON_}"  || return $?
}

function prepare_raspbian_show_list_all {
  local __keys__=
  __keys__="$( prepare_raspbian_show_list )" || return $?
  local -r __keys__

  local __json__='{}'
  local __key__=
  local __value__=
  local -i __rc__=0
  while read -r __key__ ; do
    if [ -n "${__key__}" ] ; then
      __rc__=0
      __value__="$( prepare_raspbian_check_and_get_value "${__key__}" )" || __rc__=$?
      case "${__rc__}" in
        0)
          __json__="$( jq -c -s '.[0] + .[1]' <<<"${__json__}${__value__}" )" || return $?
          ;;

        10) # Ignore (missing entry)
          __json__="$( jq -c -s '.[0] + .[1]' <<<"${__json__}{\"${__key__}\": null}" )" || return $?
          ;;

        *)
          return "${__rc__}"
          ;;
      esac
    fi
  done <<<"${__keys__}"

  jq -S '.' <<<"${__json__}" || return $?
}

function prepare_raspbian_config_check_var_name {
  local -r var_name="$1"

  local type=
  type="$( prepare_raspbian_get_type "${var_name}" )" || return $?

  if [ "${type}" = 'null' ] ; then
    echofatal "Wrong config entry. What is '${var_name}' ?"
    return 11
  fi

  return 0
}

function ensure_raspbian_config_ready {
  if [ ! -d "${__PREPARE_RASPBIAN_CONFIG_DIR__}" ] ; then
    mkdir -vp "${__PREPARE_RASPBIAN_CONFIG_DIR__}" || true
  fi
  exit_if_directory_not_exists "${__PREPARE_RASPBIAN_CONFIG_DIR__}"

  local config_file=
  config_file="$( prepare_raspbian_get_config_filename )" || return $?

  if [ ! -f "${config_file}" ] ; then
    echo '{}' >"${config_file}" # Create an empty JSON file
  fi
  exit_if_file_not_exists "${config_file}"
}

function prepare_raspbian_initialise_if_needed {
  local -r var_name="$1"
  local -r default_value="${2:-}"

# ensure_raspbian_config_ready
  local _conf_filename_=
  _conf_filename_="$( prepare_raspbian_get_config_filename )" || return $?
  local -r _conf_filename_

  local _conf_json_=
  if [ -f "${_conf_filename_}" ] ; then
    _conf_json_="$( jq -c '.' "${_conf_filename_}" )" || return $?
  else
    _conf_json_='{}'
  fi
  local -r _conf_json_

  if ! is_variable_define_and_not_empty_safe "${var_name}" ; then
    local json_config_for_var=
    json_config_for_var="$(
        jq ".[\"${var_name}\"]" <<<"${_conf_json_}"
      )" || return $?

    local value=
    if [ "${json_config_for_var}" != 'null' ] ; then
      value="$( jq -r '.value' <<<"${json_config_for_var}" )" || return $?

      if [ "${value}" = 'null' ] ; then
        local link=
        link="$( jq -r '.link' <<<"${json_config_for_var}" | envsubst )" || return $?

        if [ "${link}" != 'null' ] ; then
          if [ -f "${link}" ] ; then
            value="$( <"${link}" )"
          else
            echofatal "'${var_name}' link not valid: '${link}'"
            exit 1
          fi
        else
          value=
        fi
      fi
    fi

    declare -g "${var_name}"="${value}"

    if ! is_variable_define_and_not_empty_safe "${var_name}" ; then
      if [ -n "${default_value}" ] ; then
        echowarn "${var_name} is not define" >&2
        declare "${var_name}"="${default_value}"
        echowarn "Use ${var_name}=${!var_name}" >&2
      fi
    fi

    if is_variable_define_and_not_empty_safe "${var_name}" ; then
      declare -gr "${var_name}=${!var_name}"
      # shellcheck disable=SC2163
      export "${var_name}"
    fi
  fi
}

function _prepare_raspbian_config_json_set_ {
  local -r var_name="$1"
  local -r var_name_type="$2"
  local -r var_value="$3"

  ensure_raspbian_config_ready

  local config_file=
  config_file="$( prepare_raspbian_get_config_filename )" || return $?
  local -r config_file

  local config=
  config="$(
      jq -S ". | del(.[\"${var_name}\"] | .value) | del(.[\"${var_name}\"] | .link )" "${config_file}"
      )" || return $?

  if [ -n "${var_value}" ] ; then
    local var_value_json=

    if [ "${var_name_type}" != 'link' ] ; then
      local var_value_expected_type=
      var_value_expected_type="$( prepare_raspbian_get_type "${var_name}" )" || return $?

      case "${var_value_expected_type}" in
        'boolean'|'number')
          var_value_json="${var_value}"
          ;;

        *)
         var_value_json="\"${var_value}\""
          ;;
      esac
    else
      var_value_json="\"${var_value}\""
    fi

    config="$(
      jq -S ". |
        (.[\"${var_name}\"] | .${var_name_type} ) |= ${var_value_json}" <<<"${config}"
      )" || return $?

    echo "${config}" >"${config_file}" || return $?
    echoinfo "'${var_name}'.'${var_name_type}' is set : '${var_value}'" >&2
  else
    echo "${config}" >"${config_file}" || return $?
    echoinfo "'${var_name}' is removed." >&2
  fi
}

function _prepare_raspbian_config_set_value_ {
  local -r var_name="$1"
  local -r var_value="$2"

  _prepare_raspbian_config_json_set_ "${var_name}" 'value' "${var_value}" || return $?
}

function prepare_raspbian_check_and_set_value_for_jq_type {
  local -r var_name="$1"
  local -r var_value="$2"

  prepare_raspbian_config_check_var_name "${var_name}" || return $?
  _prepare_raspbian_config_set_value_ "${var_name}" "${var_value}" || return $?
}

function _prepare_raspbian_config_set_link_ {
  local -r var_name="$1"
  local -r var_link="$2"

  ensure_raspbian_config_ready

  _prepare_raspbian_config_json_set_ "${var_name}" 'link' "${var_link}" || return $?
}

function prepare_raspbian_check_and_set_value_link {
  local -r var_name="$1"
  local -r var_link="$2"

  prepare_raspbian_config_check_var_name "${var_name}" || return $?
 _prepare_raspbian_config_set_link_ "${var_name}" "${var_link}" || return $?
}

function prepare_raspbian_check_and_set_value {
  local -r var_type="$1"
  local -r var_name="$2"
  local -r var_value="$3"

  prepare_raspbian_config_check_var_name "${var_name}" || return $?

  local var_value_expected_type=
  var_value_expected_type="$( prepare_raspbian_get_type "${var_name}" )" || return $?

  if [ -z "${var_value_expected_type}" ] ; then
    echofatal "Don't know how to handle '${var_name}' (Internal Error)"
    return 20
  fi
  if [ "${var_type}" != 'LINK' ] ; then
    if [ "${var_type}" != "${var_value_expected_type}" ] ; then
      echofatal "'${var_name}' must be use type: '${var_value_expected_type}' (${var_type})/(${var_value})."
      return 21
    fi
  fi

  case "${var_type}" in
    'boolean') # jq type
      if ! is_boolean "${var_value}" ; then
        echofatal "Value must be 'true' or 'false': '${var_value}'"
        return 22
      fi
      prepare_raspbian_check_and_set_value_for_jq_type "${var_name}" "${var_value}" || return $?
      ;;

    'number') # jq type
      if ! is_positive_integer "${var_value}" ; then
        echofatal "Value must be a postive integer: '${var_value}'"
        return 23
      fi
      prepare_raspbian_check_and_set_value_for_jq_type "${var_name}" "${var_value}" || return $?
      ;;

    'LINK') # NOT jq type
      if [ ! -f "${var_value}" ] ; then
        echofatal "File not found: '${var_value}'"
        return 24
      fi
      prepare_raspbian_check_and_set_value_link "${var_name}" "${var_value}" || return $?
      ;;

    'string') # jq type
      prepare_raspbian_check_and_set_value_for_jq_type "${var_name}" "${var_value}" || return $?
      ;;

    *)
      echofatal "Unsupported variable type: '${var_type}' (Internal error)"
      return 123
      ;;
  esac
}

function __prepare_raspbian_check_and_get_value_for_array__ {
  local -r _name_="$1"
  local -r _values_="$2"

  local json=
  json="$( jq -c --null-input --arg k "${_name_}" '[{"key":$k,"value":[]}]' )" || return $?

  local line=
  while read -r line ; do
    if [ -n "${line}" ] ; then
      json="$( jq -c --arg v "${line}" '.[0].value += [ $v ]' <<<"${json}" )" || return $?
    fi
  done <<<"${_values_}"

  jq -c 'from_entries'  <<<"${json}" || return $?
}

function prepare_raspbian_check_and_get_value {
  local -r var_name="$1"

  prepare_raspbian_config_check_var_name "${var_name}" || return $?
  prepare_raspbian_initialise_if_needed "${var_name}" || return $?

  if is_variable_define_and_not_empty_safe "${var_name}" ; then
    local -r value="${!var_name}"
    local -i lines=
    lines="$( wc -l <<<"${value}" )"

    echodebug "var_name='${var_name}'"
    echodebug "value='${value}'"
    echodebug "lines='${lines}'"

    if [ "${lines}" -gt 1 ] ; then
      __prepare_raspbian_check_and_get_value_for_array__ "${var_name}" "${value}" || return $?
    else
      jq -c --null-input \
        --arg k "${var_name}" \
        --arg v "${value}"\
        '[{"key":$k,"value":$v}] | from_entries' || return $? # JSON format is a good idea ?
    fi
  else
    echowarn "${var_name} not define (or empty)." >&2
    return 10
  fi
}
