#!/bin/bash -e
#
# prepare-raspbian-sd:bin/api/prepare-raspbian-customize-api.sh
#
declare -gri RASPBIAN_PI_USER_ID_DEFAULT=1000
declare -gri RASPBIAN_PI_GROUP_ID_DEFAULT=1000
declare -gr  BCM_WORKSPACE_DEFAULT=true

declare -gra FILES_TO_VERIFY_ON_BOOT_PARTITION_ARMHF=(
'config.txt'
'kernel.img'
'kernel7.img'
'start.elf'
'start_x.elf'
)

declare -gra FILES_TO_VERIFY_ON_BOOT_PARTITION_ARM64=(
'config.txt'
'kernel8.img'
'start.elf'
'start_x.elf'
)

declare -gra FILES_TO_VERIFY_ON_ROOTFS_PARTITION=(
'home/pi/.bashrc'
'home/pi/.profile'
'usr/bin/raspi-config'
)

#
# API: customize_sd
#
function customize_sd {
  # Ensure boot partition is has expected
  local boot_partition_home=
  boot_partition_home="$( get_partition_home_by_label_with_timeout 'boot' )" || return $?
  echoinfo "${FUNCNAME[0]}() - boot_partition_home=${boot_partition_home}"

  # Ensure rootfs partition is has expected
  local rootfs_partition_home=
  rootfs_partition_home="$( get_partition_home_by_label_with_timeout 'rootfs' )" || return $?
  echoinfo "${FUNCNAME[0]}() - rootfs_partition_home=${rootfs_partition_home}"

  _verify_sd_content_ "${boot_partition_home}" "${rootfs_partition_home}" || return $?

  # shellcheck disable=SC2154
  if [ "${param_dry_run}" = 'false' ] ; then
    _customize_enable_ssh_ "${boot_partition_home}" || return $?
    _customize_memory_split_ "${boot_partition_home}" || return $?
    _customize_pi_user_ "${rootfs_partition_home}/home/pi" || return $?
  fi
  sync || return $?

  #
  # umount if it looks okay
  #
  local ssh_file="${boot_partition_home}/ssh"

  if [ -f "${ssh_file}" ] ; then
    echoinfo "Umounting '${boot_partition_home}'"
    if [ "${param_dry_run}" = 'false' ] ; then
      (
        set -x # show why sudo is use for
        sudo umount "${boot_partition_home}"
      ) || return $?
      echookay "${boot_partition_home} umounted."
    fi

    echoinfo "Umounting '${rootfs_partition_home}'"
    if [ "${param_dry_run}" = 'false' ] ; then
      (
        set -x # show why sudo is use for
        sudo umount "${rootfs_partition_home}"
      ) || return $?
      echookay "${rootfs_partition_home} umounted."
    fi
  else
    echowarn "ssh_file='${ssh_file}' not found."
  fi
}

function _customize_enable_ssh_ {
  local -r boot_partition_home="$1"

  if [ ! -d "${boot_partition_home}" ] ; then
    echofatal "Cannot find directory boot_partition_home='${boot_partition_home}'"
    return 100
  fi

  prepare_raspbian_initialise_if_needed SSH_ENABLE ''

  if [ "${SSH_ENABLE:-}" = 'true' ] ; then
    echoinfo 'Enabling SSH'

    local -r ssh_file="${boot_partition_home}/ssh"

    if [ -f "${ssh_file}" ] ; then
      echoinfo "File '${ssh_file}' already exists."
    else
      touch "${ssh_file}" || return $?
      echoinfo 'Enabling SSH (done)'
    fi
  else
    echoinfo 'Enabling SSH (skip)'
  fi
}

function _customize_memory_split_ {
  local -r boot_partition_home="$1"
  local -r config_txt="${boot_partition_home}/config.txt"

  prepare_raspbian_initialise_if_needed GPU_MEM '0'

  if [ "${GPU_MEM}" = '0' ] ; then
    echowarn "${FUNCNAME[0]}(): GPU_MEM is set to ${GPU_MEM} - Ignore memory split configuration"
    return 0
  fi

  if [ ! -d "${boot_partition_home}" ] ; then
    echofatal "${FUNCNAME[0]}(): Cannot find directory boot_partition_home='${boot_partition_home}'"
    return 100
  fi

  if [ ! -f "${config_txt}" ] ; then
    echofatal "${FUNCNAME[0]}(): Cannot find file config_txt='${config_txt}'"
    return 100
  fi

  if grep --quiet '^gpu_mem=' "${config_txt}" ; then
    echowarn "${FUNCNAME[0]}(): Memory split is already define - '${config_txt}'"
    grep '^gpu_mem=' "${config_txt}" >&2
    return 0
  fi

  echoinfo "${FUNCNAME[0]}(): Define memory split to minimum ('${GPU_MEM}' MB) - '${config_txt}'"
  cat <<EOF >> "${config_txt}"
# refs:
#   - https://elinux.org/RPiconfig
#   - https://www.raspberrypi.org/documentation/configuration/config-txt/memory.md
# gpu_mem=16 is the minimum possible value.
# maximum value is 192, 448, or 944, depending on whether you are using a 256M, 512MB, 1024MB or more Pi. The default value is 64.
# gpu_mem minimum GPU memory for camera use: gpu_mem=128
gpu_mem=${GPU_MEM}
EOF
}

function _customize_pi_user_ {
  local -r home_pi="$1"

  prepare_raspbian_initialise_if_needed RASPBIAN_PI_USER_ID  "${RASPBIAN_PI_USER_ID_DEFAULT}"
  prepare_raspbian_initialise_if_needed RASPBIAN_PI_GROUP_ID "${RASPBIAN_PI_GROUP_ID_DEFAULT}"
  prepare_raspbian_initialise_if_needed BCM_WORKSPACE        "${BCM_WORKSPACE_DEFAULT}"

  if [ ! -d "${home_pi}" ] ; then
    echofatal "${FUNCNAME[0]}() - Cannot find directory: /home/pi='${home_pi}'"
    return 100
  fi

  # Create bcm workspace directory (TODO: should be optional)
  if [ "${BCM_WORKSPACE:-}" != 'false' ] ; then
    local -r bcm_workspace_location="${home_pi}/.bcm-workspace"

    if [ -d "${bcm_workspace_location}" ] ; then
      echoinfo "${FUNCNAME[0]}() - Folder '${bcm_workspace_location}' already exists."
    else
      (
        set -x # show why sudo is use for
        sudo mkdir -v "${bcm_workspace_location}"
      ) || return $?
      # Set user right to 'pi' user and 'pi' group
      (
        set -x # show why sudo is use for
        sudo chown "${RASPBIAN_PI_USER_ID}:${RASPBIAN_PI_GROUP_ID}" "${bcm_workspace_location}"
      ) || return $?
    fi
  else
    echookay "${FUNCNAME[0]}() - Skip bcm-workspace. BCM_WORKSPACE='${BCM_WORKSPACE:-}'"
  fi

  # Configure ssh
  prepare_raspbian_initialise_if_needed SSH_CONFIGURE ''

  if [ "${SSH_CONFIGURE:-}" = 'true' ] ; then
    local -r ssh_folder="${home_pi}/.ssh"

    if [ -d "${ssh_folder}" ] ; then
      echoinfo "${FUNCNAME[0]}() - Folder '${ssh_folder}' already exists."
    else
      (
        set -x # show why sudo is use for
        sudo mkdir -v "${ssh_folder}"
      ) || return $?

      # Set user right to 'pi' user and 'pi' group
      (
        set -x # show why sudo is use for
        sudo chown "${RASPBIAN_PI_USER_ID}:${RASPBIAN_PI_GROUP_ID}" "${ssh_folder}"
      ) || return $?
      (
        set -x # show why sudo is use for
        sudo chmod 700 "${ssh_folder}"
      ) || return $?
    fi

    prepare_raspbian_initialise_if_needed SSH_AUTHORIZED_KEYS ''

    if ! is_variable_define_and_not_empty_safe 'SSH_AUTHORIZED_KEYS' ; then
      echowarn 'SSH_AUTHORIZED_KEYS not define (or empty) - you can fix this in file define by RASPBIAN_SSH_CONFIG'

      if is_variable_define_and_not_empty_safe 'RASPBIAN_SSH_CONFIG' ; then
        echoerror "${FUNCNAME[0]}() - Check content of '${RASPBIAN_SSH_CONFIG}'"
      fi
    else
      if sudo [ -f "${ssh_folder}/authorized_keys" ] ; then
        echowarn "${FUNCNAME[0]}() - File '${ssh_folder}/authorized_keys' already exists (nothing will be changed)."
      else
        cat <<EOF | sudo tee "${ssh_folder}/authorized_keys" || return $?
${SSH_AUTHORIZED_KEYS}
EOF
        # Set user right to 'pi' user and 'pi' group
        (
          set -x # show why sudo is use for
          sudo chown "${RASPBIAN_PI_USER_ID}:${RASPBIAN_PI_GROUP_ID}" "${ssh_folder}/authorized_keys"
        ) || return $?
        echoinfo "${FUNCNAME[0]}() - File '${ssh_folder}/authorized_keys' created."
      fi
    fi
  else
    echookay "${FUNCNAME[0]}() - Skip ~/.ssh configuration : SSH_CONFIGURE='${SSH_CONFIGURE:-}'"
  fi
}

#
# API: get_partition_home_by_label_with_timeout
#
function get_partition_home_by_label_with_timeout {
  local -r label="$1"

  local partition_home=
  for count in $(seq 10 -1 1) ; do
    if partition_home="$( _get_partition_home_by_label_ "${label}" )" ; then
      echo "${partition_home}"
      return 0
    fi
    echowarn "'${label}' partition not found. Will try again ${count} times."
    sleep 2
  done

  echofatal "'${label}' partition not found."
  return 100
}

function _get_partition_home_by_label_ {
  local -r label="$1"

  local partition_home=
  partition_home="$( mount --show-labels )" || return $?
  partition_home="$( grep "\\[${label}\\]" <<<"${partition_home}" | cut -d ' ' -f3 )" || true

  if [ -z "${partition_home}" ] ; then
    return 1
  else
    echoinfo "${FUNCNAME[0]}( '${label}' ) - ${partition_home}"
    echo "${partition_home}"
  fi
}

function _check_if_all_files_exists_ {
  local -r root_dir="$1"
  shift

  local file=
  while [ "$#" -gt 0 ] ; do
    file="${root_dir}/$1"
    shift
    if [ ! -f "${file}" ] ; then
      echowarn "File '${file}' is required."
      return 1
    fi
  done
}

function _verify_sd_content_ {
  local -r boot_partition_home="$1"
  local -r rootfs_partition_home="$2"

  if [ ! -d "${boot_partition_home}" ] ; then
    echoerror "Cannot find directory: boot_partition_home='${boot_partition_home}'"
    return 100
  fi

  if ! _check_if_all_files_exists_ "${boot_partition_home}" "${FILES_TO_VERIFY_ON_BOOT_PARTITION_ARMHF[@]}" ; then
    if ! _check_if_all_files_exists_ "${boot_partition_home}" "${FILES_TO_VERIFY_ON_BOOT_PARTITION_ARM64[@]}" ; then
      echofatal 'Missing required file'
      return 100
    fi
  fi

  if [ ! -d "${rootfs_partition_home}" ] ; then
    echoerror "Cannot find directory: rootfs_partition_home='${rootfs_partition_home}'"
    return 100
  fi

  if ! _check_if_all_files_exists_ "${rootfs_partition_home}" "${FILES_TO_VERIFY_ON_ROOTFS_PARTITION[@]}" ; then
    echofatal 'Missing required file'
    return 100
  fi
}
