#!/bin/bash
#
# prepare-raspbian-sd:bin/prepare-libreelec-sd.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -gr PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-config-json-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/libreelec-api.sh" || exit $?

function usage {
  cat <<EOF
Prepare SD card with LibreElec

Usage: $0

   $( basename "$0" ) --help
   $( basename "$0" ) --arch ARCH [OPTIONS]...

Actions:

  -h,       --help          Display help message and exit.

            --arch ARCH     Define machine hardware name:
                            --arch armhf : Download Raspberry Pi 2, Raspberry Pi 3 version
                            --arch arm32 : Same as 'armhf'
                            --arch arm64 : Download Raspberry Pi 4 version

Options:

  --download-to DIR     Upload a new image if necessary in the given directory. If the option value is
                        omitted the value defined by RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY will be used.
                        RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY:
                        '$( print_variable RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY )'

  --dev-name DEVNAME    Forces the name of the device to be used for the SD card.
                        Default value: '$( print_variable DEFAULT_SD_DEV_NAME )'

  --expected-img IMG    Expected image file name (no path). Mandatory with action --url does not download
                        the remote file if a local file is present.

EOF
}

#
# Convert arch in [ 'armhf', 'arm32', 'arm64' ]
# to hardware_type in ['RPi2.arm','RPi4.arm']
#
function convert_arch_to_hardware_type {
  local -r arch="$1" # 'armhf' / 'arm32' / 'arm64'

  local hardware_type= # 'RPi2.arm' / 'RPi4.arm'
  case "${arch}" in
    'armhf'|'arm32')
      echo 'RPi2.arm'
      return 0
      ;;

    'arm64')
      echo 'RPi4.arm'
      return 0
      ;;

    '')
      echofatal 'Parameter --arch is required'
      usage >&2
      return 1
      ;;

    *)
      echofatal "--arch not valid: '${arch}'"
      usage >&2
      return 100
      ;;
  esac
}

function getLastLibreElecUrl {
  local -r arch="$1" # 'armhf' / 'arm32' / 'arm64'

  local hardware_type=
  hardware_type="$( convert_arch_to_hardware_type "${arch}" )" || return $?
  local -r hardware_type # 'RPi2.arm' / 'RPi4.arm'

  local last_version_json=
  last_version_json="$( get_libreelec_last_image_json "${hardware_type}" )" || return $?
  local -r last_version_json

  local name=
  name="$( get_libreelec_name_from_json "${last_version_json}" )" || return $?
  local -r name

  get_libreelec_url "${name}" || return $?
}

function prepare_libreelec_sd {
  local -a params=()

  local url=
  url="$( getLastLibreElecUrl "${param_arch}" )" || return $?

  params+=(--unzip)
  params+=(--url)
  params+=("${url}")
  echoinfo "${FUNCNAME[0]}(): --url '${url}'"

  if [ -z "${param_expected_img}" ] ; then
    param_expected_img="$( basename "${url%.gz}" )" || return $?
  fi

  params+=(--expected-img)
  params+=("${param_expected_img}")
  echoinfo "${FUNCNAME[0]}(): --expected-img '${param_expected_img}'"

  if [ -n "${param_download_to}" ] ; then
    params+=(--download-to)
    params+=("${param_download_to}")
    echoinfo "${FUNCNAME[0]}(): --download-to '${param_download_to}'"
  fi

  if [ -n "${param_dev_name}" ] ; then
    params+=(--dev-name)
    params+=("${param_dev_name}")
    echoinfo "${FUNCNAME[0]}(): --dev-name '${param_dev_name}'"
  fi

  (
    set -x
    "${PROJECT_HOME}/bin/prepare-raspbian-sd.sh" "${params[@]}"
  ) || return $?
}

function main {
  local param=
  local param_arch=
  local param_dev_name=
  local param_download_to=
  local param_expected_img=

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        exit 0
        ;;

      '--download-to')
        param_download_to="$( test_P "$@" )" || return $?
        shift
        ;;

      '--dev-name')
        param_dev_name="$( test_P "$@" )" || return $?
        shift
        ;;

      '--expected-img')
        param_expected_img="$( test_P "$@" )" || return $?
        shift
        ;;

      '--arch')
        param_arch="$( test_P "$@" )" || return $?
        shift
        ;;

      *)
        echoerror "Unknown option: '${param}'"
        usage >&2
        exit 1
        ;;
    esac
  done

  prepare_libreelec_sd || return $?
}

prepare_raspbian_initialise_if_needed DEFAULT_SD_DEV_NAME '' || exit $?
prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY "$( xdg-user-dir DOWNLOAD )" || exit $?

main "$@" || exit $?

