#!/bin/bash -e
#
# prepare-raspbian-sd:bin/prepare-raspbian-sd-download-image.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -gr PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-config-json-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/raspberrypi-org-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/raspios-download-api.sh" || exit $?

declare -g ZIP_CHECKED='false'

function usage {
  cat <<EOF
Download Raspberry Pi OS image (or Raspbian) from 'raspberrypi.org' web site.

Usage
  $( basename "$0" ) --help
  $( basename "$0" ) --download [--full] [--code-name NAME] [--unzip] [--arch ARCH]
  $( basename "$0" ) --download --url IMGURL [--expected-img IMG] [--unzip]

Actions
  -h, --help            Display help message and exit.
  -D, --download        Download new image if needed into directory define by
                        RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY:
                            '$( print_variable RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY )' (Current value)
  -T DIR, --download-to DIR
                        Download new image if needed into given directory
  --url IMGURL          Allow to set image file URL

Parameters
  --arch ARCH           Define machine hardware name (optional). If not set, or empty, use "Raspbian"
                        as source, if set use "Raspberry Pi OS" as source). ARCH could be an empty string,
                        armhf or arm64.
                          --arch '' : Download last available version of raspbian
                          --arch armhf: Download last 32 bits version of "Raspberry Pi OS"
                          --arch arm32: Synonym for 'armhf'
                          --arch armv7l: Synonym for 'armv7l'
                          --arch arm64: Download last 64 bits version of "Raspberry Pi OS"
                          --arch aarch64: Synonym for 'arm64'
                        (Parameter is ignored if --url is used)
  --desktop,            Same as --full (deprecated)
  --expected-img IMG    Expected image filename (without path) - if specified when using --url allow
                        to not download remote file if local file is present.
  --full                Download full version (desktop version) instead of lite version.
                        (Parameter is ignored if --url is used)
  --code-name NAME      Force os name for raspios (buster,bullseye)
  --unzip               Unzip file after downloading
  --json                Use JSON for output using.

EOF
}

function download_if_needed {
  if is_download_needed ; then
    download || return $?
  fi
}

function is_download_needed {
  exit_if_variable_not_define_or_empty param_download_image_directory
  exit_if_directory_not_exists "${param_download_image_directory}"

  if [ -f "${EXPECTED_ZIP_FILE}" ] ; then
    echookay "File '${EXPECTED_ZIP_FILE}' already exist. (download not required)"
    ls -la --color "${EXPECTED_ZIP_FILE}" >&2
    return 1
  else
    echoinfo 'Update required.'
    return 0
  fi
}

function download {
  wget "${LAST_VERSION_URL}" -O "${EXPECTED_ZIP_FILE}" || return $?

  if [ -n "${DOWNLOAD_SHA256_URL:-}" ] ; then
    wget "${DOWNLOAD_SHA256_URL}" -O "${DOWNLOAD_SHA256_FILE}" || return $?
  fi
}

function unzip_if_needed {
  if is_unzip_needed ; then
    pushd "${param_download_image_directory}" >/dev/null || return $?

    if [ "${EXPECTED_ZIP_FILE%*.zip}.zip" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        unzip "${EXPECTED_ZIP_FILE}"
      ) || return $?
    elif [ "${EXPECTED_ZIP_FILE%*.gz}.gz" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        gzip --decompress --keep "${EXPECTED_ZIP_FILE}"
      ) || return $?
    elif [ "${EXPECTED_ZIP_FILE%*.xz}.xz" = "${EXPECTED_ZIP_FILE}" ] ; then
      (
        set -x
        xz --decompress --keep "${EXPECTED_ZIP_FILE}"
      ) || return $?
    else
      echowarn "Don't kown how to unzip: '${EXPECTED_ZIP_FILE}'" >&2
    fi
    popd >/dev/null || return $?
  fi
  return 0
}

function is_unzip_needed {
  exit_if_variable_not_define_or_empty param_download_image_directory
  exit_if_directory_not_exists "${param_download_image_directory}"

  exit_if_file_not_exists "${EXPECTED_ZIP_FILE}"

  local target_img="${EXPECTED_IMG_FILE}"

  if [ -f "${target_img}" ] ; then
    echookay "File '${target_img}' already exist. (unzip not required)"
    ls -la --color "${target_img}" >&2
    return 1
  else
    echoinfo "unzip required -> '${target_img}'"
    return 0
  fi
}

function check_zip_once {
  if [ "${ZIP_CHECKED}" = 'true' ] ; then
    return 0
  fi

  if [ ! -f "${EXPECTED_ZIP_FILE}.sha256" ] ; then
    echofatal "Cannot find '${EXPECTED_ZIP_FILE}.sha256' - Cannot trust zip file."
    return 100
  fi

  echoinfo "Verifying checksum of '${EXPECTED_ZIP_FILE}'" >&2
  local check256=
  check256="$(
    cd "$( dirname "${EXPECTED_ZIP_FILE}" )" || exit $?
    sha256sum "$( basename "${EXPECTED_ZIP_FILE}" )"
    )" || return $?

  local verify256=
  verify256="$( cat "${EXPECTED_ZIP_FILE}.sha256" )" || return $?

  if [ "${check256}" != "${verify256}" ] ; then
    echofatal "Wrong sha256sum !"
    cat <<EOF >&2
COMPUTED: ${check256}
EXPECTED: ${verify256}
EOF
    return 100
  fi

  ZIP_CHECKED='true'
}

function main {
  local action=
  local param=
  local param_arch=             # 'armhf'|'arm64'|'' also 'arm32' alias for 'armhf'
  local param_download_image_directory=
  local param_expected_img=
  local param_output_json=
  local param_code_name=        # 'buster' or 'bullseye'
  local param_os_type='lite'    # 'full'|'lite'
  local param_unzip=
  local param_url=

  exit_if_command_is_missing \
    curl \
    gzip \
    sha256sum \
    unzip \
    wget

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage
        return 0
        ;;

      '-D'|'--download')
        action='DOWNLOAD'
        exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY'
        param_download_image_directory="${RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY}"
        ;;

      '-T'|'--download-to')
        action='DOWNLOAD'
        param_download_image_directory="$( P_param "$@" )" && shift || return $?
        ;;

      '--arch')
        param_arch="$( P_param "$@" )" && shift || return $?
        ;;

      '--desktop'|'--full')
        param_os_type='full'
        ;;

      '--os-name'|'--code-name') # --os-name is deprecated
        param_code_name="$( P_param "$@" )" && shift || return $?
        ;;

      '--unzip')
          param_unzip='true'
          ;;

      '--json')
          param_output_json='true'
          ;;

      '--url')
        param_url='true'
        LAST_VERSION_URL="$( P_param "$@" )" && shift || return $?
        declare -gr LAST_VERSION_URL
        ;;

      '--expected-img')
        param_expected_img="$( P_param "$@" )" && shift || return $?
        ;;

      *)
        echoerror "Unknown option: '${param}'"
        usage >&2
        return 1
        ;;
    esac
  done

  if [ -z "${action}" ] ; then
    echofatal 'Action is missing.'
    usage >&2
    return 1
  fi

  if [ -z "${param_download_image_directory}" ] ; then
    exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY'
    param_download_image_directory="${RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY}"
  fi

  case "${param_arch}" in
    '')
      # OK, User want formerly raspbian (probably no more supported in the futur)
      ;;
    'armhf'|'arm32'|'armv7l')
       param_arch='armhf' # OK, user want armhf
       ;;
    'arm64'|'aarch64')
       param_arch='arm64' # OK, user want arm64
       ;;
    *)
      echofatal "--arch '${param_arch}' is not valid. '', 'armhf' or 'arm64' is expected."
      return 1
      ;;
  esac

  case "${param_os_type}" in
    'full'|'lite') ;; # OK
    '') param_os_type='lite' ;; # OK

    *)
      echofatal "--os-type '${param_os_type}' is not valid. 'full' or 'lite' is expected."
      return 1
      ;;
  esac

  if [ -z "${param_url}" ] ; then
    raspios_download_init_variables "${param_code_name}" "${param_arch}" "${param_os_type}" "${param_download_image_directory}" || return $?
  else
    local basename=
    basename="$( basename "${LAST_VERSION_URL}" )"

    if [ -n "${param_expected_img}" ] ; then
      EXPECTED_IMG_FILE="${param_download_image_directory}/${param_expected_img}"
    else
      EXPECTED_IMG_FILE=
    fi

    declare -gr DOWNLOAD_SHA256_FILE= # empty
    declare -gr DOWNLOAD_SHA256_URL= # empty
    declare -gr EXPECTED_IMG_FILE
    declare -gr EXPECTED_ZIP_FILE="${param_download_image_directory}/${basename}"
    declare -gr LAST_VERSION_FILENAME= # empty

    raspios_download_show_variables
  fi

  case "${action}" in
    'DOWNLOAD')
      download_if_needed || return $?

      if [ -z "${param_url}" ] ; then
        check_zip_once || return $?
      else
        echowarn "Cannot trust downloaded file (don't how to get hashcode)."
      fi
      ;;

    *)
      echofatal "Action not valid: '${action}'"
      usage >&2
      return 1
      ;;
  esac

  if [ "${param_unzip}" = 'true' ] ; then
    if [ -z "${param_url}" ] ; then
      check_zip_once || return $?
    else
      echowarn "Cannot trust downloaded file (don't how to get hashcode)."
    fi

    unzip_if_needed || return $?
  fi

  if [ "${param_output_json}" = 'true' ] ; then
    jq --null-input \
      --arg arch "${EXPECTED_ZIP_FILE}" \
      --arg img "${EXPECTED_IMG_FILE}" \
      --arg url "${LAST_VERSION_URL}" \
      --arg file "${LAST_VERSION_FILENAME:-__UNDEFINED__}" \
      '{arch: $arch, img: $img, url: $url, file: $file }' || return $?
  else
    cat <<EOF
LAST_VERSION_URL     :${LAST_VERSION_URL}
LAST_VERSION_FILENAME:${LAST_VERSION_FILENAME:-__UNDEFINED__}
EXPECTED_ZIP_FILE    :${EXPECTED_ZIP_FILE}
EXPECTED_IMG_FILE    :${EXPECTED_IMG_FILE}
LAST_ZIP_FILE        :${EXPECTED_ZIP_FILE}
LAST_IMG_FILE        :${EXPECTED_IMG_FILE}
EOF
  fi
}

prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY "$( xdg-user-dir DOWNLOAD )"

main "$@" || exit $?
