#!/bin/bash
#
# prepare-raspbian-sd:bin/prepare-raspbian-sd.sh
#
set -u
set -o pipefail

PROJECT_HOME="$( dirname "$( dirname "$( realpath "${BASH_SOURCE[0]}" )" )" )"
declare -gr PROJECT_HOME

# shellcheck disable=SC1091
source "${PROJECT_HOME}/apis/bash-common-api/bash-common-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-config-json-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-customize-api.sh" || exit $?
# shellcheck disable=SC1091
source "${PROJECT_HOME}/bin/api/prepare-raspbian-detect-sd-api.sh" || exit $?

function usage {
  cat <<EOF
Puts the image of the Raspberry Pi (formerly Raspbian) operating system on an SD card
and allows it to be configured. Also allows you to install images of which you know the
URL. Also allows you to create an image from an SD card.

Usage

   $( basename "$0" ) --help
   $( basename "$0" ) ACTION [OPTIONS]...

Actions

  -h,       --help          Display help message and exit.
  -I FILE,  --put-image FILE
                            Initialize the SD card using the FILE image file.
  -D,       --put-default-image
                            Initialize the SD card using the defined image by RASPBIAN_DEFAULT_IMG:
                            '$( print_variable RASPBIAN_DEFAULT_IMG )'
  -L,       --put-last-image
                            Initialize the SD card using the latest available version of the Raspbian image.
                            Download the image from the Raspbian website if necessary.
            --configure     Apply the user configuration on the SD card.
  -F,       --full-clone    Create an SD card image in the directory defined by
                            RASPBIAN_DEFAULT_BACKUP_DIRECTORY (a file name will be generated from the current date).
                            RASPBIAN_DEFAULT_BACKUP_DIRECTORY:
                            '$( print_variable RASPBIAN_DEFAULT_BACKUP_DIRECTORY )'
  -T FILE,  --full-clone-to FILE
                            Create an image of the SD card using given name. FILE could be a relative or
                            absolute filename. Fail if path does not exists or file already exists.
  --compute-sd-dev-name     Locate the SD card and display the names of associated devices, then exit.

Options

  --arch ARCH           Define machine hardware name (optional). If not set, or empty, use "Raspbian"
                        as source, if set use "Raspberry Pi OS" as source). ARCH could be an empty string,
                        armhf or arm64.
                          --arch '' : Download last available version of raspbian
                          --arch armhf: Download last 32 bits version of "Raspberry Pi OS"
                          --arch arm32: Same as 'armhf'
                          --arch arm64: Download last 64 bits version of "Raspberry Pi OS"
  --arm32               Shortcut for --arch arm32
  --arm64               Shortcut for --arch arm64
  --os-name NAME        Force os name for raspios (buster,bullseye)
  --auto-umount, --auto-unmount
                        Auto umount device (Warning use with --dry-run first)
  --debug               Show more traces
  --desktop             Same as --full (deprecated)
  --dev-name DEVNAME    Forces the name of the device to be used for the SD card.
                        Default value: '$( print_variable DEFAULT_SD_DEV_NAME )'
  --download-to DIR     Upload a new image if necessary in the given directory. If the option value is
                        omitted the value defined by RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY will be used.
                        RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY:
                        '$( print_variable RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY )'
  --dry-run             Simulate ACTION.
  --expected-img IMG    Expected image file name (no path). Mandatory with action --url does not download
                        the remote file if a local file is present.
                        (If use with --unzip must be unzip name of file)
  --full                Download full version (the desktop version) of Raspberry Pi OS instead of the
                        light version (which is server version).
  --gui                 Same as --full (deprecated)
  --unzip               Unzip file after downloading, and use unzipped file as image file
                        (gzip and unzip supported)

EOF

  if [ "${1:-}" = '--help' ] ; then
    cat <<EOF
Security:
  To write to SD (except for the action --configure), the tool checks that the SD card is NOT mounted/
  You must be running umount (or using the interface graphic)
    sudo umount /dev/<DEVNAME>

Samples:

  prepare-raspbian-sd -L --arch armhf # Server 32 bits version
  prepare-raspbian-sd -L --arch armhf --desktop # Desktop 32 bits version
  prepare-raspbian-sd -L --arch arm64 # Server 64 bits version
  prepare-raspbian-sd -L --arch arm64 --desktop # Desktop 64 bits version

    Initialize tour SD card with last Rasperry Pi OS image (SD must be plugin but not mounted)

  prepare-raspbian-sd -L

    Initialize your SD card with last Raspbian image (SD must be plugin but not mounted)

  prepare-raspbian-sd -T ~/my-sd-backup.img

    Create an image of your SD named ~/my-sd-backup.img

  prepare-raspbian-sd -F

    Create an image of your SD, name is generated and use folder define by RASPBIAN_DEFAULT_BACKUP_DIRECTORY
    According to your configuration:
      '$( print_variable RASPBIAN_DEFAULT_BACKUP_DIRECTORY )'

EOF
  fi
}

function run_prepapre_raspbian_sd_download_image {
  local __output__=
  __output__="$(
    if [ "${DEBUG:-}" = 'true' ] ; then
      set -x
    fi
    "${PROJECT_HOME}/bin/prepare-raspbian-sd-download-image.sh" --json "$@"
    )" || return $?

# grep 'LAST_IMG_FILE:' <<<"${__output__}" | cut -d ':' -f2 || return $?
  jq -r '.img' <<<"${__output__}" || return $?
}

function get_last_img_version {
  local -a options=()

  options+=(--unzip)
  options+=(--download-to "${download_to}")

  [ -n "${param_full}" ] && options+=("${param_full}")
  [ -n "${param_arch}" ] && options+=(--arch "${param_arch}")
  [ -n "${param_code_name}" ] && options+=(--code-name "${param_code_name}")

  run_prepapre_raspbian_sd_download_image "${options[@]}" || return $?
}

function download_img_file {
  if [ -z "${param_url}" ] ; then
    echofatal 'Value for --url is missing.'
    return 100
  fi
  if [ -z "${param_expected_img}" ] ; then
    echofatal 'Value for --expected-img is missing.'
    return 100
  fi

  local -a options=()

  options+=(--download-to "${download_to}")
  options+=(--url "${param_url}")
  options+=(--expected-img "${param_expected_img}")

  [ "${param_unzip}" = 'true' ] && options+=('--unzip')

  run_prepapre_raspbian_sd_download_image "${options[@]}" || return $?
}

function get_default_backup_file {
  echo "${RASPBIAN_DEFAULT_BACKUP_DIRECTORY}/$( date +'%Y-%m-%d.%H-%M' )-SD.img"
}

function ask_to_umout_or_auto_umount {
  local -r sd_full_dev_name="$1"

  if mount | grep "^${sd_full_dev_name}" ; then

    local dev_list=
    dev_list="$( mount | grep "${sd_full_dev_name}" | cut -d ' ' -f1 | sort )" || return $?
    local -r dev_list

    if [ "${param_auto_umount}" = 'true' ] ; then
      local dev=
      while read -r dev ; do
        if [ -n "${dev}" ] ; then
          echo "Try to umount '${dev}'"
          (
            set -x # show why sudo is use for
            sudo umount "${dev}"
          ) || return $?
        fi
      done <<<"${dev_list}"

    else

      echofatal "SD Card '${sd_full_dev_name}' is mounted, you need to run:"

      local dev=
      while read -r dev ; do
        if [ -n "${dev}" ] ; then
          echo "sudo umount ${dev}"
        fi
      done <<<"${dev_list}"
      return 100
    fi

  fi
}

function erase_sd_card_with_image {
  local -r sd_full_dev_name="$1"
  local -r image_file="$2"

  if [ ! -f "${image_file}" ] ; then
    echoerror "Image file '${image_file}' not found."
    return 100
  fi

  ask_to_umout_or_auto_umount "${sd_full_dev_name}" || return $?

  echoinfo "Prepare SD based on: '${image_file}'"

  # Cancel ?
  if ! whiptail --yesno --defaultno "Would you like to erase content of '${sd_full_dev_name}' ?" 7 60 2 ; then # not yes
    echoinfo "Cancel by user."
    exit 1
  fi

  local -i _dd_rc_=0
  date
  if [ "${param_dry_run}" = 'false' ] ; then
    (
      set -x # show why sudo is use for
      sudo dd bs=4M if="${image_file}" of="${sd_full_dev_name}" status=progress conv=fsync
    ) || _dd_rc_=$?
  else
    cat <<EOF >&2
dry runs:
      sudo dd bs=4M if="${image_file}" of="${sd_full_dev_name}" status=progress conv=fsync
EOF
  fi
  sync
  date

  if [ "${_dd_rc_}" -ne 0 ] ; then
    echofatal "dd: ${_dd_rc_}"
    return "${_dd_rc_}"
  else
    echoinfo "dd: ${_dd_rc_}"
  fi
}

function create_sd_complete_image {
  local -r sd_full_dev_name="$1"
  local -r image_file="$2"

  if [ -f "${image_file}" ] ; then
    echoerror "Image file '${image_file}' already exists."
    return 100
  fi

  ask_to_umout_or_auto_umount "${sd_full_dev_name}" || return $?

  local -i _dd_rc_=0
  date
  if [ "${param_dry_run}" = 'false' ] ; then
    (
      set -x
      sudo dd bs=4M if="${sd_full_dev_name}" of="${image_file}" status=progress conv=fsync
    ) || _dd_rc_=$?
  else
    cat <<EOF >&2
dry runs:
      sudo dd bs=4M if="${sd_full_dev_name}" of="${image_file}" status=progress conv=fsync
EOF
  fi
  sync
  date
  if [ "${_dd_rc_}" -ne 0 ] ; then
    echofatal "dd: ${_dd_rc_}"
    return "${_dd_rc_}"
  else
    echoinfo "dd: ${_dd_rc_}"
  fi

  local -r currentuser="${USER}"

  if [ -f "${image_file}" ] ; then
    cat <<EOF >&2
      sudo chown -v "${currentuser}:${currentuser}" "${image_file}"
EOF
    if [ "${param_dry_run}" = 'false' ] ; then
      (
        set -x # show why sudo is use for
        sudo chown -v "${currentuser}:${currentuser}" "${image_file}"
      ) || return $?
    else
      cat <<EOF >&2
dry runs:
      sudo chown -v "${currentuser}:${currentuser}" "${image_file}"
EOF
    fi
  fi

  echoinfo "Image file '${image_file}' created."
  echoinfo "Real path: '$( realpath "${image_file}" )'"
}

function verify_sd {
  local -r sd_dev_name_only="$1"

  if [ -f "/sys/block/${sd_dev_name_only}/device/type" ] ; then
    read -r is_sd < "/sys/block/${sd_dev_name_only}/device/type"

    if [ "${is_sd}" != 'SD' ] ; then
      echofatal "Content of '/sys/block/${sd_dev_name_only}/device/type' is not 'SD' - found '${is_sd}'"
      return 100
    fi
  else
    echofatal "File not found: '/sys/block/${sd_dev_name_only}/device/type'"
    return 100
  fi
}

function get_SD_device_safe {
  local sd_dev_name=
  sd_dev_name="$( detect_sd_get_SD_device )" || true # Don't fail on error

  if [ -z "${sd_dev_name}" ] ; then
     echofatal 'No SD devices found. Do you have a connected SD card?'
  elif [ "$( wc -w <<<"${sd_dev_name}" )" != 1 ] ; then
    echofatal "Computed device for SD give more than one value '${sd_dev_name}'"
    echofatal 'Use must use --dev-name option'
  else
    echo "${sd_dev_name}"
  fi
}

function init_SD_DEV_NAME {
  if [ -n "${param_force_sd_dev_name}" ] ; then
    declare -gr SD_DEV_NAME="${param_force_sd_dev_name}"

    echoinfo "Fixed SD_DEV_NAME='${SD_DEV_NAME}'"
  elif [ "${param_compute_sd_dev_name}" = 'true' ] || ! is_variable_define_and_not_empty_safe DEFAULT_SD_DEV_NAME; then
    SD_DEV_NAME="$( get_SD_device_safe )" || true # Don't fail on error
    declare -gr SD_DEV_NAME

    if [ -z "${SD_DEV_NAME}" ] ; then
      echofatal 'Cannot compute SD_DEV_NAME'
      echofatal 'Use must use --dev-name option or plug SD Card.'
      return 101
    else
      local -r sd_full_dev_name="/dev/${SD_DEV_NAME}"

      echoinfo "Computed SD_DEV_NAME='${SD_DEV_NAME}'" >&2
      ls -l --color "${sd_full_dev_name}"* >&2
      mount | grep "${sd_full_dev_name}" | cut -d ' ' -f1 | sort || true
    fi

  else
    declare -gr SD_DEV_NAME="${DEFAULT_SD_DEV_NAME}"

    echoinfo "Default SD_DEV_NAME='${SD_DEV_NAME}'"
  fi
}

function main {
  local action=
  local download_to=
  local param=
  local param_arch=
  local param_auto_umount=
  local param_compute_sd_dev_name='false'
  local param_dry_run='false'
  local param_expected_img=
  local param_force_sd_dev_name=
  local param_full=
  local param_image_file=
  local param_code_name=
  local param_unzip=
  local param_url=

  exit_if_command_is_missing \
    curl \
    dd \
    gzip \
    mount \
    umount \
    unzip \
    whiptail

  while [ "$#" -gt 0 ] ; do
    param="$1"
    shift

    case "${param}" in
      '-h'|'--help')
        usage --help
        exit 0
        ;;

      '-D'|'--put-default-image')
        action='WRITE_ON_SD'
        exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_IMG'
        param_image_file="${RASPBIAN_DEFAULT_IMG}"
        ;;

      '-I'|'--put-image')
        action='WRITE_ON_SD'
        param_image_file="$( P_param "$@" )" && shift || return $?
        ;;

      '-L'|'--put-last-image')
        action='WRITE_ON_SD'
        param_image_file=':LAST:'
        ;;

      '--url')
        action='WRITE_ON_SD'
        param_image_file=':URL:'
        param_url="$( P_param "$@" )" && shift || return $?
        ;;

      '--configure')
        action='CONFIGURE'
        ;;

      '-F'|'--full-clone')
        action='READ_FROM_SD'
        exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_BACKUP_DIRECTORY'
        param_image_file="$( get_default_backup_file )"
        ;;

      '-T'|'--full-clone-to')
        action='READ_FROM_SD'
        param_image_file="$( P_param "$@" )" && shift || return $?
        ;;

      '--debug')
        declare -gx DEBUG='true'
        declare -gx MESSAGE_API_SHOW_DEBUG='true'
        ;;

      '--dry-run')
        param_dry_run='true'
        ;;

      '--arch')
        param_arch="$( P_param "$@" )" && shift || return $?
        ;;

      '--arm64')
        param_arch='arm64'
        ;;

      '--arm32')
        param_arch='arm32'
        ;;

      '--os-name'|'--code-name') # --os-name is deprecated
        param_code_name="$( P_param "$@" )" && shift || return $?
        ;;

      '--auto-umount'|'--auto-unmount')
        param_auto_umount='true'
        ;;

      '--download-to')
        download_to="$( P_param "$@" )" && shift || return $?
        ;;

      '--desktop'|'--full'|'--gui')
        param_full='--full'
        ;;

      '--dev-name')
        param_force_sd_dev_name="$( P_param "$@" )" && shift || return $?
        ;;

      '--compute-sd-dev-name')
        action='DRYRUN'
        param_compute_sd_dev_name='true'
        ;;

      '--expected-img')
        param_expected_img="$( P_param "$@" )" && shift || return $?
        ;;

      '--unzip')
        param_unzip='true'
        ;;

      *)
        echoerror "Unknown option: '${param}'"
        usage >&2
        return 1
        ;;
    esac
  done

  if [ -z "${action}" ] ; then
    echoerror 'Action si missing.'
    usage >&2
    return 1
  fi

  if [ -z "${download_to}" ] ; then
    exit_if_variable_not_define_or_empty 'RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY'
    download_to="${RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY}"
  fi
  if [ "${param_image_file}" = ':LAST:' ] ; then
    param_image_file="$( get_last_img_version )"

    if [ -z "${param_image_file}" ] ; then
      echofatal 'Cannot get last image'
      return 100
    fi
  elif [ "${param_image_file}" = ':URL:' ] ; then
    param_image_file="$( download_img_file )" || return $?

    if [ -z "${param_image_file}" ] ; then
      echofatal "Cannot get '${param_image_file}'"
      return 100
    fi
  fi

  init_SD_DEV_NAME || return $?

  local -r sd_dev_name_only="${SD_DEV_NAME}"
  local -r sd_full_dev_name="/dev/${SD_DEV_NAME}"

  verify_sd "${sd_dev_name_only}" || return $?

  case "${action}" in
    'CONFIGURE')
      customize_sd || return $?
      ;;

    'WRITE_ON_SD')
      erase_sd_card_with_image "${sd_full_dev_name}" "${param_image_file}" || return $?
      ;;

    'READ_FROM_SD')
      create_sd_complete_image "${sd_full_dev_name}" "${param_image_file}" || return $?
      ;;

    'DRYRUN')
      # --compute-sd-dev-name
      ;;

    *)
      echofatal "Action not valid (${action}')"
      usage >&2
      exit 1
      ;;
  esac
}

prepare_raspbian_initialise_if_needed DEFAULT_SD_DEV_NAME '' || exit $?
prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_BACKUP_DIRECTORY "$( xdg-user-dir DOCUMENTS )" || exit $?
prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_DOWNLOAD_IMAGE_DIRECTORY "$( xdg-user-dir DOWNLOAD )" || exit $?
prepare_raspbian_initialise_if_needed RASPBIAN_DEFAULT_IMG '' || exit $?

main "$@" || exit $?

